(* eval.sml *)

(* evaluation of abstract syntax expressions, declarations, and statements in an
 * environment *)

structure Eval = 
struct

open Exprs
open Env

(* These are the values computed by eval and bound in environments.
 * The three variants of this datatype determine the "dynamically
 * checkable" types. Dynamically, all functions are of the same type. *)
datatype value
  = IntVal of int               (* integers *)
  | BoolVal of bool             (* booleans *)
  | FunVal of funExpr * venv    (* function closure *)

(* venv: value environments *)
withtype venv = value env

(* note the use of a "withtype" clause to allow a type abbreviation
 * to be associated with a datatype declaration.  The value datatype
 * indirectly self-recursive via the venv abbreviation. *)

(* "values" of statements *)
datatype stmtValue
  = ExprVal of value
  | DeclVal of variable * value

(* the following evaluation functions do "dynamic type checking" by
 * examining the intermediate values and checking that they have the
 * expected type (i.e. int, bool, or function) *)

(* EvalTypeError is raised with a dynamic type check fails *)
exception EvalTypeError

(* BadOp raised when a unary operator is applied as a binary op,
 * or vice versa *)
exception BadOp

(* applyUnOp : operator * value -> value *)
fun applyUnOp (oper, v) =
    (case oper
       of Neg  =>
	    (case v
	      of IntVal v1 => IntVal (~v1)
	       | _ => raise EvalTypeError)
        | Not  =>
	    (case v
	      of BoolVal v1 => BoolVal (not v1)
	       | _ => raise EvalTypeError)
        | _ => raise BadOp)

fun applyBinOp (oper, e1, e2, env) =
    let fun checkIntArgs (opr) =
	    case (eval(e1, env), eval(e2, env))
	     of (IntVal n1, IntVal n2) => opr (n1,n2)
	      | _ => raise EvalTypeError
    in case oper
	 of Plus =>
	      IntVal(checkIntArgs (op +))
	  | Minus =>
	      IntVal(checkIntArgs (op -))
	  | Times =>
	      IntVal(checkIntArgs (op * ))
	  | Div =>
	      IntVal(checkIntArgs (op div))
	  | Mod =>
	      IntVal(checkIntArgs (op mod))
	  | Equal =>
	      BoolVal(checkIntArgs (op =))
	  | NotEqual =>
	      BoolVal(checkIntArgs (op <>))
	  | Less =>
	      BoolVal(checkIntArgs (op <))
	  | LessEq =>
	      BoolVal(checkIntArgs (op <=))
	  | Greater =>
	      BoolVal(checkIntArgs (op >))
	  | GreaterEq =>
	      BoolVal(checkIntArgs (op >=))
          (* we implement short-circuit evaluation of And and Or *)
	  | And => 
	      (case eval(e1, env)
                 of BoolVal b1 => 
                      if b1 then 
                        (case eval(e2, env)
                           of v2 as BoolVal _ => v2
                            | _ => raise EvalTypeError)
		      else BoolVal false
		  | _ =>  raise EvalTypeError)
	  | Or => 
	      (case eval(e1, env)
                 of v1 as BoolVal b1 => 
                      if b1 then v1
                      else 
                        (case eval(e2, env)
                           of v2 as BoolVal _ => v2
                            | _ => raise EvalTypeError)
		  | _ =>  raise EvalTypeError)
          | _ => raise BadOp
    end (* fun applyBinOp *)

(* eval : expr * venv -> int *)
and eval (e, env) =
    case e
      of Nat n => IntVal n
       | True => BoolVal true
       | False => BoolVal false
       | Id v => lookup (v, env)
       | UnaryApp(oper, e1) =>
	   applyUnOp(oper, eval(e1, env))
       | BinaryApp(oper, e1, e2) =>
	   applyBinOp(oper, e1, e2, env)
       | If (e1,e2,e3) =>
           (* we could check that the values of e2 and e3 have
            * the same dynamic type, but we don't *)
	   (case eval(e1, env)
	      of BoolVal b =>
		   if b then eval(e2, env) else eval(e3, env)
	       | _ => raise EvalTypeError)
       | App (f, arg) =>
	   (case lookup(f,env)
             of funval as FunVal(Fn(fname,param,body), envClosure) =>
		  let val argval = eval(arg, env)
		      val env' = bind(fname, funval,
                                   bind(param, argval, envClosure))
		  in eval(body, env')
		  end
	      | _ => raise EvalTypeError)


(* evalDecl : venv -> decl -> stmtValue * venv *)
fun evalDecl env (Let(v,e)) = 
    let val def = eval(e, env)
     in (DeclVal(v,def), bind(v, def, env))
    end
  | evalDecl env (Fun(f,x,e)) = 
    let val fval = FunVal(Fn(f,x,e), env)
     in (DeclVal(f,fval), bind(f, fval, env))
    end

(* evalStmt : env -> stmt -> (stmtValue * env) *)
fun evalStmt env (Decl d) = evalDecl env d
  | evalStmt env (Expr e) = (ExprVal(eval(e, env)), env)

end (* structure Eval *)


(* Notes

This evaluator is fairly typical of a simple interpreter.  It takes
an expression (or declaration/statement) and an environment mapping
variables (strings) to values.  The environment is needed to interpret
free variables that occur in expressions.

In this simple calculator language, variables are only bound at the
"top level" -- there are no local binding mechanisms except for the
parameter variable binding in function declarations.

The eval functions do dynamic type checking, making sure that the
values passed to primitive functions have the right types, and raising
an exception in case they do not.  This sort of dynamic checking is
not available for applications of user-definded functions.  A dynamic
type error aborts the evaluation by raising the EvalTypeError exception.
With more effort, more detailed error reporting could be implemented
(e.g. reporting which operator was being applied, and what the erroneous
argument type 

A static type checker could eliminate the possibility of these dynamically
checked type errors, and then we could simplify the evaluator by removing
the checks.

The evaluator supports recursive functions.  For instance, the factorial
function is defined by

    fun f x = if x == 0 then 1 else x * f(x-1)

Recursive calls are supported in the "App (f, arg)" case of eval by adding
the function binding, as well as the parameter variable binding, to the
closure environment when evaluating the function body.

*)
